{# database_structure_tree.html #}

{% extends "base.html" %}

{% block content %}

    <h1>Research database structure: tree form</h1>

    {% if with_database %}
        <p>Default database: {{ default_database }}</p>
    {% endif %}
    <p>Default schema: {{ default_schema }}</p>

    {% include "show_hide_buttons.html" %}

    {{ content | safe }}

{% endblock %}
