{# contact_after_discharge_permission_phrase.html #}

{% comment %}
    template parameters:
        permitted_to_contact_discharged_patients_for_n_days: int
        permitted_to_contact_discharged_patients_for_n_years: str
{% endcomment %}

CPFT has ethics approval to contact patients for up
to {{ permitted_to_contact_discharged_patients_for_n_days }} days
({{ permitted_to_contact_discharged_patients_for_n_years }} years)
after discharge, without specific consent.
