# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-12 01:37
from __future__ import unicode_literals

import crate_anon.common.jsonfunc
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('userprofile', '0005_userprofile_patient_multiquery_scratchpad'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userprofile',
            name='patient_multiquery_scratchpad',
            field=crate_anon.common.jsonfunc.JsonClassField(null=True, verbose_name='PatientMultiQuery scratchpad (in JSON) for builder'),  # noqa
        ),
    ]
