{# phrase_yellow.html #}

<em><b>YELLOW. I’m not sure yet. Please ask my CPFT clinicians to consider
any requests from research teams and pass them on to me if appropriate.</b>
I will choose each time whether researchers can have access to my CPFT
records and write to me.</em>
