{# phrase_green.html #}

<em><b>GREEN. I am happy to hear directly from approved CPFT researchers inside
or outside my clinical team. Please allow these researchers access to my full
CPFT records and allow them to write to me about projects that might be suited
to me.</b> I understand that there’s no commitment to take part in any
future project.</em>
