{# query_missing.html #}

{% extends "base.html" %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>Failed: no query</h1>
    <p>There is no active query. Try creating one.</p>

{% endblock %}
