{# pe_bad.html #}

{% extends "base.html" %}

{% block extrahead %}
    <style type="text/css">{{ sql_highlight_css }}</style>
{% endblock %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>Failed: error in Patient Explorer</h1>
    <p>There’s an error in one of your Patient Explorer queries.
        Please edit it and try again.</p>

    <h1>Error</h1>
    <div class="code">{{ exception }}</div>

    <h1>Query that failed</h1>
    {{ query|safe }}

{% endblock %}
