<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01//EN"
    "http://www.w3.org/TR/html4/strict.dtd">
<html lang="en">
    <!-- 500.html -->
    <head>
        <title>Page unavailable</title>
    </head>
    <body>
        <h1>Page unavailable</h1>
    
        <p>Sorry, but the requested page is unavailable due to a
        server hiccup.</p>
    
        <p>Our engineers have been notified, so check back later.</p>
    </body>
</html>
