{# query_none_selected.html #}

{% extends "base.html" %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>No query selected</h1>
    <p>Please create or choose a query and try again.</p>

{% endblock %}
